use color_eyre::Result;
use ratatui::layout::Constraint;
use ratatui::{
    text::Text,
    layout::{Alignment, Rect}, 
    widgets::{Block, Cell, Row, Table}, 
    Frame
};

use super::{PartialComponent, Component, CurrentTab};
use crate::{
    action::Action, 
    config::Config, 
    palette::StylePalette, 
};

pub struct StakingDetails {
    palette: StylePalette,
    staked_own: u128,
    staked_total: u128,
}

impl Default for StakingDetails {
    fn default() -> Self {
        Self::new()
    }
}

impl StakingDetails {
    const TICKER: &str = " CSPR";
    const DECIMALS: usize = 5;

    pub fn new() -> Self {
        Self {
            palette: StylePalette::default(),
            staked_own: 0,
            staked_total: 0,
        }
    }

    fn prepare_u128(&self, value: u128) -> String {
        let value = value as f64 / 10f64.powi(18);
        let after = Self::DECIMALS;
        format!("{:.after$}{}", value, Self::TICKER)
    }
}

impl PartialComponent for StakingDetails {
    fn set_active(&mut self, _current_tab: CurrentTab) { }
}

impl Component for StakingDetails {
    fn register_config_handler(&mut self, config: Config) -> Result<()> {
        if let Some(style) = config.styles.get(&crate::app::Mode::Validator) {
            self.palette.with_normal_style(style.get("normal_style").copied());
            self.palette.with_hover_style(style.get("hover_style").copied());
            self.palette.with_normal_border_style(style.get("normal_border_style").copied());
            self.palette.with_hover_border_style(style.get("hover_border_style").copied());
            self.palette.with_normal_title_style(style.get("normal_title_style").copied());
            self.palette.with_hover_title_style(style.get("hover_title_style").copied());
            self.palette.with_highlight_style(style.get("highlight_style").copied());
            self.palette.with_scrollbar_style(style.get("scrollbar_style").copied());
        }
        Ok(())
    }

    fn update(&mut self, action: Action) -> Result<Option<Action>> {
        match action {
            Action::SetStakedRatio(total, own) => {
                self.staked_total = total;
                self.staked_own = own;
            }
            _ => {}
        };
        Ok(None)
    }

    fn draw(&mut self, frame: &mut Frame, area: Rect) -> Result<()> {
        let [_, place, _] = super::validator_balance_layout(area);
        let (border_style, border_type) = self.palette.create_border_style(false);

        let table = Table::new(
            vec![
                Row::new(vec![
                    Cell::from(Text::from("Stake value".to_string()).alignment(Alignment::Left)),
                    Cell::from(Text::from(self.prepare_u128(self.staked_total)).alignment(Alignment::Right)), 
                ]),
                Row::new(vec![
                    Cell::from(Text::from("Own stake".to_string()).alignment(Alignment::Left)),
                    Cell::from(Text::from(self.prepare_u128(self.staked_own)).alignment(Alignment::Right)), 
                ]),
                Row::new(vec![
                    Cell::from(Text::from("Other stake".to_string()).alignment(Alignment::Left)),
                    Cell::from(Text::from(self.prepare_u128(self.staked_total.saturating_sub(self.staked_own))).alignment(Alignment::Right)), 
                ]),
            ],
            [
                Constraint::Min(11),
                Constraint::Min(0),
            ],
        )
        .highlight_style(self.palette.create_highlight_style())
        .column_spacing(1)
        .block(Block::bordered()
            .border_style(border_style)
            .border_type(border_type)
            .title_alignment(Alignment::Right)
            .title_style(self.palette.create_title_style(false))
            .title("Staking details"));

        frame.render_widget(table, place);

        Ok(())
    }
}
