use color_eyre::Result;
use subxt::{
    ext::sp_core::{sr25519::Pair, Pair as PairT},
    tx::{PairSigner, TxProgress}, 
    OnlineClient,
};
use tokio::sync::mpsc::UnboundedSender;

use crate::{
    action::Action, 
    casper::{CasperConfig, CasperExtrinsicParamsBuilder}, 
    casper_network::{self, runtime_types}, 
    types::{ActionLevel, ActionTarget},
};

pub async fn transfer_balance(
    action_tx: &UnboundedSender<Action>,
    api: &OnlineClient<CasperConfig>,
    sender: &[u8; 32],
    receiver: &[u8; 32],
    amount: &u128,
    mut maybe_nonce: Option<&mut u32>,
) -> Result<TxProgress<CasperConfig, OnlineClient<CasperConfig>>> {
    let receiver_id = subxt::utils::MultiAddress::Id(
        subxt::utils::AccountId32::from(*receiver)
    );

    let transfer_tx = casper_network::tx()
        .balances()
        .transfer_allow_death(receiver_id, *amount);

    let tx_params = match maybe_nonce {
        Some(ref mut nonce) => {
            **nonce = nonce.saturating_add(1);
            CasperExtrinsicParamsBuilder::new()
                .nonce(nonce.saturating_sub(1) as u64)
                .build()
        },
        None => CasperExtrinsicParamsBuilder::new().build(),
    };

    let pair = Pair::from_seed(sender);
    let signer = PairSigner::<CasperConfig, Pair>::new(pair);

    match api
        .tx()
        .sign_and_submit_then_watch(&transfer_tx, &signer, tx_params)
        .await {
            Ok(tx_progress) => {
                action_tx.send(Action::EventLog(
                        format!("transfer transaction {} sent", tx_progress.extrinsic_hash()),
                        ActionLevel::Info,
                        ActionTarget::WalletLog))?;
                Ok(tx_progress)
            },
            Err(err) => {
                action_tx.send(Action::EventLog(
                        format!("error during transfer: {err}"), 
                        ActionLevel::Error,
                        ActionTarget::WalletLog))?;
                Err(err.into())
            }
        }
}

pub async fn bond_extra(
    action_tx: &UnboundedSender<Action>,
    api: &OnlineClient<CasperConfig>,
    sender: &[u8; 32],
    amount: &u128,
    mut maybe_nonce: Option<&mut u32>,
) -> Result<TxProgress<CasperConfig, OnlineClient<CasperConfig>>> {
    let transfer_tx = casper_network::tx()
        .staking()
        .bond_extra(*amount);

    let tx_params = match maybe_nonce {
        Some(ref mut nonce) => {
            **nonce = nonce.saturating_add(1);
            CasperExtrinsicParamsBuilder::new()
                .nonce(nonce.saturating_sub(1) as u64)
                .build()
        },
        None => CasperExtrinsicParamsBuilder::new().build(),
    };

    let pair = Pair::from_seed(sender);
    let signer = PairSigner::<CasperConfig, Pair>::new(pair);

    match api
        .tx()
        .sign_and_submit_then_watch(&transfer_tx, &signer, tx_params)
        .await {
            Ok(tx_progress) => {
                action_tx.send(Action::EventLog(
                        format!("bond extra transaction {} sent", tx_progress.extrinsic_hash()),
                        ActionLevel::Info,
                        ActionTarget::ValidatorLog))?;
                Ok(tx_progress)
            },
            Err(err) => {
                action_tx.send(Action::EventLog(
                        format!("error during bond extra: {err}"), 
                        ActionLevel::Error,
                        ActionTarget::ValidatorLog))?;
                Err(err.into())
            }
        }
}

pub async fn bond(
    action_tx: &UnboundedSender<Action>,
    api: &OnlineClient<CasperConfig>,
    sender: &[u8; 32],
    amount: &u128,
    mut maybe_nonce: Option<&mut u32>,
) -> Result<TxProgress<CasperConfig, OnlineClient<CasperConfig>>> {
    // auto-stake everything by now
    let reward_destination = 
        casper_network::runtime_types::pallet_staking::RewardDestination::Staked;

    let transfer_tx = casper_network::tx()
        .staking()
        .bond(*amount, reward_destination);

    let tx_params = match maybe_nonce {
        Some(ref mut nonce) => {
            **nonce = nonce.saturating_add(1);
            CasperExtrinsicParamsBuilder::new()
                .nonce(nonce.saturating_sub(1) as u64)
                .build()
        },
        None => CasperExtrinsicParamsBuilder::new().build(),
    };

    let pair = Pair::from_seed(sender);
    let signer = PairSigner::<CasperConfig, Pair>::new(pair);

    match api
        .tx()
        .sign_and_submit_then_watch(&transfer_tx, &signer, tx_params)
        .await {
            Ok(tx_progress) => {
                action_tx.send(Action::EventLog(
                        format!("bond transaction {} sent", tx_progress.extrinsic_hash()),
                        ActionLevel::Info,
                        ActionTarget::ValidatorLog))?;
                Ok(tx_progress)
            },
            Err(err) => {
                action_tx.send(Action::EventLog(
                        format!("error during bond: {err}"), 
                        ActionLevel::Error,
                        ActionTarget::ValidatorLog))?;
                Err(err.into())
            }
        }
}

pub async fn payout_stakers(
    action_tx: &UnboundedSender<Action>,
    api: &OnlineClient<CasperConfig>,
    sender: &[u8; 32],
    stash: &[u8; 32],
    era_index: u32,
    mut maybe_nonce: Option<&mut u32>,
) -> Result<TxProgress<CasperConfig, OnlineClient<CasperConfig>>> {
    let stash_id = subxt::utils::AccountId32::from(*stash);
    let transfer_tx = casper_network::tx()
        .staking()
        .payout_stakers(stash_id, era_index);

    let tx_params = match maybe_nonce {
        Some(ref mut nonce) => {
            **nonce = nonce.saturating_add(1);
            CasperExtrinsicParamsBuilder::new()
                .nonce(nonce.saturating_sub(1) as u64)
                .build()
        },
        None => CasperExtrinsicParamsBuilder::new().build(),
    };

    let pair = Pair::from_seed(sender);
    let signer = PairSigner::<CasperConfig, Pair>::new(pair);

    match api
        .tx()
        .sign_and_submit_then_watch(&transfer_tx, &signer, tx_params)
        .await {
            Ok(tx_progress) => {
                action_tx.send(Action::EventLog(
                        format!("payout stakers {} sent", tx_progress.extrinsic_hash()),
                        ActionLevel::Info,
                        ActionTarget::ValidatorLog))?;
                Ok(tx_progress)
            },
            Err(err) => {
                action_tx.send(Action::EventLog(
                        format!("error during payout stakers: {err}"), 
                        ActionLevel::Error,
                        ActionTarget::ValidatorLog))?;
                Err(err.into())
            }
        }
}

pub async fn set_keys(
    action_tx: &UnboundedSender<Action>,
    api: &OnlineClient<CasperConfig>,
    sender: &[u8; 32],
    hashed_keys_str: &String,
    mut maybe_nonce: Option<&mut u32>,
) -> Result<TxProgress<CasperConfig, OnlineClient<CasperConfig>>> {
    let (gran_key, babe_key, audi_key, slow_key) = {
        let s = hashed_keys_str.trim_start_matches("0x");
        (
            hex::decode(&s[0..64]).unwrap().as_slice().try_into().unwrap(),
            hex::decode(&s[64..128]).unwrap().as_slice().try_into().unwrap(),
            hex::decode(&s[128..192]).unwrap().as_slice().try_into().unwrap(),
            hex::decode(&s[192..256]).unwrap().as_slice().try_into().unwrap(),
        )
    };

    let session_keys = runtime_types::casper_runtime::opaque::SessionKeys {
        grandpa: runtime_types::sp_consensus_grandpa::app::Public(gran_key),
        babe: runtime_types::sp_consensus_babe::app::Public(babe_key),
        authority_discovery: runtime_types::sp_authority_discovery::app::Public(audi_key),
        slow_clap: runtime_types::ghost_slow_clap::sr25519::app_sr25519::Public(slow_key),
    };

    // it seems like there is no check for the second paramter, that's why
    // we it can be anything.
    // Not sure TBH
    let transfer_tx = casper_network::tx()
        .session()
        .set_keys(session_keys, Vec::new());

    let tx_params = match maybe_nonce {
        Some(ref mut nonce) => {
            **nonce = nonce.saturating_add(1);
            CasperExtrinsicParamsBuilder::new()
                .nonce(nonce.saturating_sub(1) as u64)
                .build()
        },
        None => CasperExtrinsicParamsBuilder::new().build(),
    };

    let pair = Pair::from_seed(sender);
    let signer = PairSigner::<CasperConfig, Pair>::new(pair);

    match api
        .tx()
        .sign_and_submit_then_watch(&transfer_tx, &signer, tx_params)
        .await {
            Ok(tx_progress) => {
                action_tx.send(Action::EventLog(
                        format!("set keys {} sent", tx_progress.extrinsic_hash()),
                        ActionLevel::Info,
                        ActionTarget::ValidatorLog))?;
                Ok(tx_progress)
            },
            Err(err) => {
                action_tx.send(Action::EventLog(
                        format!("error during set keys: {err}"), 
                        ActionLevel::Error,
                        ActionTarget::ValidatorLog))?;
                Err(err.into())
            }
        }
}

pub async fn validate(
    action_tx: &UnboundedSender<Action>,
    api: &OnlineClient<CasperConfig>,
    sender: &[u8; 32],
    percent: u32,
    mut maybe_nonce: Option<&mut u32>,
) -> Result<TxProgress<CasperConfig, OnlineClient<CasperConfig>>> {
    let validator_prefs = casper_network::runtime_types::pallet_staking::ValidatorPrefs {
        commission: runtime_types::sp_arithmetic::per_things::Perbill(percent),
        blocked: percent >= 1_000_000_000u32,
    };

    let transfer_tx = casper_network::tx()
        .staking()
        .validate(validator_prefs);

    let tx_params = match maybe_nonce {
        Some(ref mut nonce) => {
            **nonce = nonce.saturating_add(1);
            CasperExtrinsicParamsBuilder::new()
                .nonce(nonce.saturating_sub(1) as u64)
                .build()
        },
        None => CasperExtrinsicParamsBuilder::new().build(),
    };

    let pair = Pair::from_seed(sender);
    let signer = PairSigner::<CasperConfig, Pair>::new(pair);

    match api
        .tx()
        .sign_and_submit_then_watch(&transfer_tx, &signer, tx_params)
        .await {
            Ok(tx_progress) => {
                action_tx.send(Action::EventLog(
                        format!("set keys {} sent", tx_progress.extrinsic_hash()),
                        ActionLevel::Info,
                        ActionTarget::ValidatorLog))?;
                Ok(tx_progress)
            },
            Err(err) => {
                action_tx.send(Action::EventLog(
                        format!("error during set keys: {err}"), 
                        ActionLevel::Error,
                        ActionTarget::ValidatorLog))?;
                Err(err.into())
            }
        }
}
