use std::path::PathBuf;
use std::fs::File;
use std::io::{Write, BufRead, BufReader};

use color_eyre::Result;

#[derive(Debug)]
pub struct LocalStorage<ValueType> {
    file_path: PathBuf,
    values: Vec<(String, ValueType)>,
}

impl<ValueType: std::fmt::Display> LocalStorage<ValueType> {
    pub fn new(file_path: PathBuf) -> Self {
        Self { file_path, values: Vec::new() }
    }

    fn read_or_create<F1, F2>(
        &mut self, 
        make_values: F1, 
        fallback_on_error: F2,
    ) -> Result<()>
    where
        F1: Fn(&[u8; 32]) -> ValueType,
        F2: Fn(&PathBuf) -> Vec<(String, ValueType)>,
    {
        assert!(self.values.len() == 0, "values already imported from local storage");
        match File::open(&self.file_path) {
            Ok(file) => {
                let reader = BufReader::new(file);

                for line in reader.lines() {
                    let line = line?;

                    let line_split_at = line
                        .find(":")
                        .unwrap_or(line.len() - 1);

                    let (name, value) = line.split_at(line_split_at);
                    let value = &value[3..value.len() - 1];

                    let value: [u8; 32] = hex::decode(value)
                        .expect("stored seed is valid hex string; qed")
                        .as_slice()
                        .try_into()
                        .expect("stored seed is valid length; qed");

                    self.values.push((name.to_string(), make_values(&value)));
                }
            },
            Err(_) => self.values = fallback_on_error(&self.file_path),
        };
        self.write_to_file()?;
        Ok(())
    }

    fn write_to_file(&self) -> Result<()> {
        let mut file = File::create(&self.file_path)?;
        self.values
            .iter()
            .map(|value| {
                writeln!(file, "{}:{}", value.0, value.1)
                    .expect("should write in new file; qed");
            });
        Ok(())
    }
}
